#define IN_INASH_C
#include <unistd.h>
#include "inash.h"  /* stdio.h and stdlib.h */
#include <setjmp.h>
#include <errno.h>
#include <stdarg.h>

#define FPTBL_SIZE 3
#define BLOCK_OUT -2  /* for doformat */
#define OUTPUT_ERR 01 /* for doformat */

static void nullfunc();
static int fprintf_inash_va(FILE *fp, const char *format, va_list va);
static char *malloc_and_format(const char *format,va_list va);

static FILE *fptbl[FPTBL_SIZE];
static jmp_buf jbuf;
static int retcode;
static void (*exitfunc)(void);

void nullfunc(void){
  return;
}

int start_inash(){
  int count;
  for(count=0; count<FPTBL_SIZE; count++)
    fptbl[count]=NULL;
  exitfunc=&nullfunc;
  return 0;  /* reserved */
}

int origmain_inash(mainfunc *origmain, int argc, char *argv[]){
  if(!setjmp(jbuf)) retcode=(*origmain)(argc,argv);
  return retcode;
}

int end_inash(){
  fcloseall_inash();
  return 0; /* reserved */
}

void exit_inash(int r){
  retcode=r;
  longjmp(jbuf,1);
}

void atexit_inash(void (*func)(void) ){
  exitfunc=func;
  return;
}

FILE *fopen_inash(const char *file, const char *mode){
  int count;

  for(count=0; count<FPTBL_SIZE; count++){
    if(fptbl[count]==NULL){
      /* register the file pointer */
      return fptbl[count]=fopen(file,mode);
    }
  }
  /* if fptbl is full */
  errno=EMFILE;   /* "Too many open file" error */
  return NULL;
}

int fclose_inash(FILE *fp){
  int count;

  if(fp==NULL) return 0;
  if(fp==stdout){
    flushout(out1);
    return 0;
  }
  if(fp==stderr){
    flushout(out2);
    return 0;
  }
  if(fp==stdin) return 0;

  for(count=0; count<FPTBL_SIZE; count++){
    if(fptbl[count]==fp){
      /* if the file has been opened via fopen_inash */
      fptbl[count]=NULL;
      return fclose(fp);
    }
  }
  /* if the file pointer is not found in fptbl,
     simply close the file.                     */
  return fclose(fp);
}

int fcloseall_inash(void){
  int count;
  for(count=0; count<FPTBL_SIZE; count++){
    if(fptbl[count]!=NULL){
      fclose_inash(fptbl[count]);
    }
  }
}

int fflush_inash(FILE *fp){
  int retcode=0;
  int count;
  
  if(fp==stdout){
    flushout(out1);
    return 0;
  }
  if(fp==stderr){
    flushout(out2);
    return 0;
  }
  if(fp==NULL){         /* flush all files */
    /* if at least one error occured, return EOF. */
    flushout(out1);
    flushout(out2);
    for(count=0; count<FPTBL_SIZE; count++)
      if(fptbl[count]!=NULL)
        if(!fflush(fptbl[count]))
	  retcode=EOF;
    return retcode;
  }
  /* otherwise */
  return fflush(fp);
}

int fputc_inash(int ch, FILE *fp){
  if(fp==stdout){
    out1c((unsigned char)ch);
    return ch;
  }
  if(fp==stderr){
    out2c((unsigned char)ch);
    return ch;
  }
  return fputc(ch,fp);
}

int fputs_inash(const char *s, FILE *fp){
  if(fp==stdout){
    out1str(s);
    flushout(out1);
    return 0;
  }
  if(fp==stderr){
    out2str(s);
    flushout(out2);
    return 0;
  }
  return fputs(s,fp);
}

int puts_inash(const char *s){
  out1str(s);
  out1c('\n');
  flushout(out1);
  return 0;
}

int fprintf_inash(FILE *fp, const char *format, ...){
  va_list va;
  int r;

  va_start(va,format);
  r=fprintf_inash_va(fp,format,va);
  va_end(va);
  return r;
}

int printf_inash(const char *format,...){
  va_list va;
  int r;

  va_start(va,format);
  r=fprintf_inash_va(stdout,format,va);
  va_end(va);
  return r;
}

static int fprintf_inash_va(FILE *fp, const char *format, va_list va){
  char *buf;
  int retcode;

  buf=malloc_and_format(format,va);
  if(buf==NULL) return EOF;
  
  retcode=strlen(buf);
  if(fp==stdout) out1str(buf);
  else if(fp==stderr) out2str(buf);
  else if(fputs(buf,fp)!=0) retcode=EOF;
  
  free(buf);
  return retcode;
}

/* 
 * static char *malloc_and_format(const char *format, va_list va);
 * RETURNS: 
 *   pointer to the allocated local heap including formatted string
 *   NULL if an error occured (in this case, no memory allocated)
 *
 * struct output{
 *   char *nextc;
 *   int nleft;
 *   char *buf;
 *   int bufsize;
 *   short fd;
 *   short flags;
 * }
*/

static char *malloc_and_format(const char *format, va_list va){
  char *mem;
  size_t memsize;
  size_t memsizestep;
  static struct output strout;
  
  memsizestep=strlen(format);
  memsize=memsizestep*2;

  for(;;){
    mem=malloc(memsize);
    if(mem==NULL) return NULL;
    strout.nextc=mem;
    strout.nleft=memsize;
    strout.fd=BLOCK_OUT;
    strout.flags=0;
    doformat(&strout,(char *)format,va);
    outc('\0',&strout);

    if( !(strout.flags & OUTPUT_ERR) ){
      /* if memsize is large enough */
      return mem;
    }else if( ( strout.nextc>=mem )
              && ( strout.nextc<mem+memsize )
	    ){
      /* if unexpected error occured */
      free(mem);
      return NULL;
    }

    /* when memsize is too small */
    free(mem);
    memsize+=memsizestep;
  }
}

int fgetc_inash(FILE *fp){
  if(fp==stdin) return getchar_inash();
  else return fgetc(fp);
}

int getchar_inash(){
  unsigned char c;
  
  if(read(0,&c,1)!=1) return EOF;
  else return (int)c;
}

char *fgets_inash(char *buf, int n, FILE *fp){
  char *p;
  char *bufend;
  int ch;

  if(fp==stdin){
    if(n<2) return NULL;
    p=buf;
    bufend=buf+n-1;
    while(p<bufend){
      ch=getchar_inash();
      if(ch==EOF) break;
      *(p++)=(unsigned char)ch;
      if(ch=='\n') break;
    }
    if(p==buf) return NULL; /* EOF */
    *p='\0';
    return buf;
  }
  return fgets(buf,n,fp);
}

char *gets_inash(char *buf){
  char *p;
  int ch;

  ch=getchar_inash();
  if(ch==EOF) return NULL;
  p=buf;
  while(ch!=EOF || ch!='\n'){
    *(p++)=(unsigned char)ch;
    ch=getchar_inash();
  }
  *p='\0';
  return buf;
}


